/*
 * $Id: gt_stats.c,v 1.8 2003/07/21 16:48:07 hipnod Exp $
 *
 * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "gt_gnutella.h"

#include "gt_node.h"
#include "gt_netorg.h"

#include "gt_search.h"

/*****************************************************************************/

struct gt_stats
{
	double          size_kb;
	unsigned long   files;
	size_t          users;
};

/*****************************************************************************/

/* 
 * This should be computed from our neighboring peers, but for now it
 * hardcodes the max leaves for the most numerous nodes on the network,
 * Limewire (32) and BearShare (10) 
 */
static int avg_leaves     = (10 + 32) / 2;

/* default ttl we use in searches */
static int default_ttl    = GT_SEARCH_TTL;

/* default outdegree for most nodes */
static int default_degree = 6;

/*****************************************************************************/

/* no need to pull in math.h for this simple thing: exponent is small */
static unsigned long int_pow (int base, int exponent)
{
	unsigned long total = 1;

	while (exponent-- > 0)
		total *= base;

	return total;
}

/* TODO: do this on a per-node basis, and use X-MaxTTL and X-Degree */
static unsigned long sum_network (int degree, int ttl)
{
	int            i;
	unsigned long  sum;

	if (ttl <= 0)
		return 0;

	sum = degree;

	for (i = 2; i <= ttl; i++)
		sum += int_pow (degree-1, i-1) * degree;

	return sum;
}

/* NOTE: of course this is totally inaccurate, and not even dynamic */
unsigned long guess_users (size_t connected)
{
	unsigned long users;

	users = sum_network (default_degree, default_ttl) * connected;
	
	/* multiply by the number of leaves */
	users *= avg_leaves;

	/* divide by redundancy level */
	users /= 3;

	/* divide by 3 to account for cycles */
	users /= 3;

	return users;
}

/*****************************************************************************/

static TCPC *count_stats (TCPC *c, GtNode *node, struct gt_stats *st)
{
	st->size_kb += node->size_kb;
	st->files   += node->files;

	if (st->size_kb > 0 || st->files > 0)
		st->users++;

	return NULL;
}

/*
 * TODO: accumulate statistics on average leaves/cycles on our immediate
 *       peers.
 */
int gnutella_stats (Protocol *p, unsigned long *users, unsigned long *files,
                    double *size, Dataset **extra)
{
	struct gt_stats  stats;
	size_t           connected;
	unsigned long    avg_users;
	unsigned long    avg_files;
	double           avg_size;

	memset (&stats, 0, sizeof (struct gt_stats));

	*files = *users = *size = 0;

	connected = gt_conn_length (GT_NODE_ULTRA, GT_NODE_CONNECTED);

	if (connected == 0)
		return TRUE;

	gt_conn_foreach (GT_CONN_FOREACH(count_stats), &stats, GT_NODE_NONE,
	                 -1, 0);

	if (stats.users == 0)
		return TRUE;

	avg_files = stats.files / stats.users;
	avg_size  = stats.size_kb / stats.users;

	avg_users = guess_users (connected);
	*users = avg_users;

	*files = avg_files * avg_users;
	*size  = avg_size * avg_users / 1024 / 1024;

	return TRUE;
}
